﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Common;
using HeuristicLab.Core.Views;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.MainForm;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.EvolutionaryTracking.Views {
  [View("LineageExplorerView")]
  [Content(typeof(LineageExplorer), IsDefaultView = false)]
  public sealed partial class LineageExplorerView : ItemView {
    private Dictionary<TreeNode, ISymbolicExpressionTree> treeMap;
    private List<SymbolicExpressionGenealogyGraphNode> lineage;

    public new LineageExplorer Content {
      get { return (LineageExplorer)base.Content; }
      set { base.Content = value; }
    }

    public LineageExplorerView() {
      InitializeComponent();

      treeMap = new Dictionary<TreeNode, ISymbolicExpressionTree>();
    }

    protected override void DeregisterContentEvents() {
      // TODO: Deregister your event handlers here
      base.DeregisterContentEvents();
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      // TODO: Register your event handlers here
    }

    #region Event Handlers (Content)
    // TODO: Put event handlers of the content here
    #endregion

    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content == null) {
        // TODO: Add code when content has been changed and is null
      } else {
        var tuples = Content.Trees.OrderByDescending(x => x.Item2).ToList();
        scopeListTreeView.Nodes.AddRange(tuples.Select(x => new TreeNode(x.Item2.ToString())).ToArray());

        for (int i = 0; i != tuples.Count; ++i) {
          treeMap.Add(scopeListTreeView.Nodes[i], tuples[i].Item1);
          double quality = tuples[i].Item2;
          int colorIndex = (int)Math.Floor(quality * ColorGradient.Colors.Count);
          var color = ColorGradient.Colors[colorIndex];
          scopeListTreeView.Nodes[i].ForeColor = color;
        }

        // TODO: Add code when content has been changed and is not null
      }
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      // TODO: Enable or disable controls based on whether the content is null or the view is set readonly
    }

    private void scopeListTreeView_AfterSelect(object sender, TreeViewEventArgs e) {
      var treeNode = scopeListTreeView.SelectedNode;
      var symbExprTree = treeMap[treeNode];
      HighlightFragment(symbExprTree);

      var graphNode = Content.GenealogyGraph.GetGraphNodes(symbExprTree).First();
      if (lineage == null)
        lineage = new List<SymbolicExpressionGenealogyGraphNode> { graphNode };
      else
        lineage.Clear();

      var gn = graphNode;
      while (gn.InEdges != null && gn.InEdges.Count != 0) {
        gn = (SymbolicExpressionGenealogyGraphNode)gn.InEdges[0].Source;
        lineage.Add(gn);
      }
      lineage.Reverse();

      // update the righthand side tree view 
      qualityImprovementTreeView.Nodes.Clear();
      for (int i = 0; i != lineage.Count; ++i) {
        var node = new TreeNode(string.Format("Generation {0}", lineage[i].Rank));
        qualityImprovementTreeView.Nodes.Add(node);
        var quality = lineage[i].Quality;
        int colorIndex = (int)Math.Floor(quality * ColorGradient.Colors.Count);
        var color = ColorGradient.Colors[colorIndex];
        node.ForeColor = color;
      }
    }

    /// <summary>
    /// Given a symbolic expression tree, find the corresponding vertex in the genealogy graph, and highlight the fragment
    /// that is saved as a data member in the incoming arc from the parent in the previous generation
    /// </summary>
    /// <param name="tree"></param>
    private void HighlightFragment(ISymbolicExpressionTree symbExprTree) {
      // update the symbolic expression tree chart with the new selected tree:
      symbolicExpressionTreeChart.Tree = symbExprTree;
      var matchingNodes = Content.GenealogyGraph.GetGraphNodes(symbExprTree);
      SymbolicExpressionGenealogyGraphNode graphNode = matchingNodes.First();

      if (graphNode.InEdges == null) return;

      var fragment = (IFragment)graphNode.InEdges[0].Data;
      if (fragment.Root == null) return;
      foreach (var node in fragment.Root.IterateNodesBreadth()) {
        var visualNode = symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node);
        visualNode.LineColor = Color.DodgerBlue;
      }
      symbolicExpressionTreeChart.Repaint();
    }

    private void qualityImprovementTreeView_AfterSelect(object sender, TreeViewEventArgs e) {
      var tree = lineage[qualityImprovementTreeView.SelectedNode.Index].SymbolicExpressionTree;
      HighlightFragment(tree);
    }

    private void calculateFragmentFrequencyButton_Click(object sender, EventArgs e) {
      var matchingNodes = Content.GenealogyGraph.GetGraphNodes(symbolicExpressionTreeChart.Tree);
      SymbolicExpressionGenealogyGraphNode graphNode = matchingNodes.First();

      //      var tree = symbolicExpressionTreeChart.Tree;
      //      var fragments = tree.IterateNodesPrefix().Select(node => new Fragment(node)).ToList();
      //
      //      var similarityComparer = new SymbolicExpressionTreeNodeSimilarityComparer {
      //        MatchConstantValues = false,
      //        MatchVariableNames = true,
      //        MatchVariableWeights = false
      //      };
      //
      //      var dict = new Dictionary<ISymbolicExpressionTree, bool>();
      //
      //      var trees = Content.Trees.Select(t => t.Item1).ToList();


      #region oldcode
      if (graphNode.InEdges == null) return;

      var fragment = (IFragment)graphNode.InEdges[0].Data;
      if (fragment.Root == null) return;

      var fragments = new List<IFragment>(); // fragments that contain the current fragment
      var subtree = fragment.Root;

      while (subtree.Parent != null) {
        subtree = subtree.Parent;
        fragments.Add(new Fragment(subtree));
      }

      fragments.Add(fragment);
      fragments.Sort((b, a) => a.Length.CompareTo(b.Length));

      var similarityComparer = new SymbolicExpressionTreeNodeSimilarityComparer {
        MatchConstantValues = false,
        MatchVariableNames = true,
        MatchVariableWeights = false
      };

      //      var trees = Content.GenealogyGraph.Nodes.Select(n => n.SymbolicExpressionTree).ToList();

      var trees = Content.Trees.Select(t => t.Item1).ToList();
      //      var dictionary = new Dictionary<ISymbolicExpressionTree, bool>(trees.Count);
      var hashset = new HashSet<ISymbolicExpressionTree>();
      //      foreach (var tree in trees)
      //        if (!dictionary.ContainsKey(tree))
      //          dictionary.Add(tree, false);

      foreach (var f in fragments) {
        double frequency = 0;
        foreach (var tree in trees) {
          if (hashset.Contains(tree)) frequency += 1;
          else if (tree.Root.ContainsFragment(f, similarityComparer)) {
            frequency += 1;
            hashset.Add(tree);
          }
        }

        int index = (int)Math.Floor(frequency * ColorGradient.Colors.Count / trees.Count);
        if (index == ColorGradient.Colors.Count) index--;
        var color = ColorGradient.Colors[index];

        foreach (var node in f.Root.IterateNodesBreadth()) {
          var visualNode = symbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node);
          visualNode.FillColor = Color.FromArgb(200, color);
        }
      }
      #endregion
      symbolicExpressionTreeChart.Repaint();
    }

    private void symbolicExpressionTreeChart_SymbolicExpressionTreeNodeClicked(object sender, MouseEventArgs e) {

    }

    private void showGenealogyButton_Click(object sender, EventArgs e) {
      var tree = symbolicExpressionTreeChart.Tree;
      if (tree == null) return;
      var graphNode = Content.GenealogyGraph.GetGraphNodes(tree).First();

      var graph = new SymbolicExpressionTreeGenealogyGraph();
      var ancestors = graphNode.Ancestors();
      foreach (var ancestorNode in ancestors) {
        graph.AddNode(ancestorNode);
      }
      using (var dialog = new GenealogyGraphDialog { Graph = graph }) {
        //        dialog.ShowDialog();
        dialog.Show(this);
      }
    }

    #region Event Handlers (child controls)
    // TODO: Put event handlers of child controls here.
    #endregion
  }
}
